/*
    Melody maid
    copyright (c) 1998-2013 Kazuki Iwamoto http://www.maid.org/ iwm@maid.org

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "abort.h"
#include "add.h"
#include "command.h"
#include "edit.h"
#include "export.h"
#include "file.h"
#include "general.h"
#include "genre.h"
#include "mm_dialog.h"
#include "mm_list.h"
#include "mmaid.h"
#include "process.h"
#include "reload.h"
#include "root.h"
#include "sigmain.h"
#include "version.h"
#include "misc/misc.h"
#include "orz/orzcharlist.h"
#include "orz/orzcharset.h"
#include "orz/orzhistory.h"
#include "orz/orzmdi.h"
#include "orz/orzpref.h"


/******************************************************************************
*                                                                             *
* ja:メニュー関数群                                                           *
*                                                                             *
******************************************************************************/
void
command_new (GtkWidget *widget,
             gpointer   user_data)
{
  file_open_edit (NULL);
}


void
command_open (GtkWidget *widget,
              gpointer   user_data)
{
  gchar *file = NULL, *title;
  GtkStockItem stock_item;
  GtkWidget *dialog;

  gtk_stock_lookup (GTK_STOCK_OPEN, &stock_item);
  title = misc_mnemonic_to_text (stock_item.label);
#if GTK_CHECK_VERSION(2,4,0)
  dialog = gtk_file_chooser_dialog_new (title,
                        GTK_WINDOW (window), GTK_FILE_CHOOSER_ACTION_OPEN,
# ifdef G_OS_WIN32
                        GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
                        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
# else /* not G_OS_WIN32 */
                        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                        GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
# endif /* not G_OS_WIN32 */
                        NULL);
  if (open_path)
    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (dialog), open_path);
  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    file = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
#else /* not GTK_CHECK_VERSION(2,4,0) */
  dialog = gtk_file_selection_new (title);
  if (open_path)
    {
      gchar *path;

      path = g_strconcat (open_path, G_SEARCHPATH_SEPARATOR_S, NULL);
      gtk_file_selection_set_filename (GTK_FILE_SELECTION (dialog), path);
      g_free (path);
    }
  gtk_file_selection_hide_fileop_buttons (GTK_FILE_SELECTION (dialog));
  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_OK)
    file = g_strdup (gtk_file_selection_get_filename
                                                (GTK_FILE_SELECTION (dialog)));
#endif /* not GTK_CHECK_VERSION(2,4,0) */
  g_free (title);
  gtk_widget_destroy (dialog);
  if (file)
    {
      gchar *path;

      g_free (open_path);
      path = g_path_get_absolute (file, NULL);
      open_path = g_path_get_dirname (path);
      g_free (path);
      file_open_edit (file);
      g_free (file);
    }
}


void
command_close (GtkWidget *widget,
               gpointer   user_data)
{
  gint page;

  if ((page = gtk_notebook_get_current_page (GTK_NOTEBOOK (mdi))) >= 0
    && prompt_close ((MmaidWindow *)orz_mdi_get_data (ORZ_MDI (mdi), page)))
    gtk_notebook_remove_page (GTK_NOTEBOOK (mdi), page);
}


void
command_save (GtkWidget *widget,
              gpointer   user_data)
{
  gint page_num;

  page_num = gtk_notebook_get_current_page (GTK_NOTEBOOK (mdi));
  if (orz_mdi_get_created (ORZ_MDI (mdi), page_num))
    {
      command_saveas (widget, user_data);
    }
  else
    {
      gboolean result = TRUE;
      const gchar *file;
      MmaidWindow *mmaid;

      file = orz_mdi_get_file (ORZ_MDI (mdi), page_num);
      mmaid = orz_mdi_get_data (ORZ_MDI (mdi), page_num);
      if (g_file_test (file, G_FILE_TEST_EXISTS))
        {
          gchar *utf8str;
          GtkWidget *dialog;

          utf8str = g_filename_display_name (file);
          dialog = gtk_message_dialog_new (GTK_WINDOW (window),
                    GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
                    GTK_MESSAGE_INFO, GTK_BUTTONS_YES_NO,
                    _("%s\nThis file is existed.\n\nOver write?"), utf8str);
          g_free (utf8str);
          g_signal_connect (G_OBJECT (dialog), "key-press-event",
                                    G_CALLBACK (misc_dialog_key_press), NULL);
          result = gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_YES;
          gtk_widget_destroy (dialog);
        }
      if (result && (mmaid->filetype == FILE_TYPE_DB
            ? file_save_db (file, mmaid, mmaid->charset)
            : file_save_list (file, mmaid, mmaid->charset,
                mmaid->playlist, mmaid->separator, mmaid->crlf, mmaid->header))
                 && (mmaid->filetype == FILE_TYPE_DB
                                    || mmaid->playlist != PLAYLIST_BASENAME))
        orz_mdi_set_edited (ORZ_MDI (mdi), page_num, FALSE);
    }
}


typedef struct _SaveDialog
{
  GtkWidget *radio_db, *radio_plist;
  GtkWidget *frame_path, *radio_relative, *radio_absolute, *radio_basename;
  GtkWidget *frame_sep, *radio_slash, *radio_yen;
  GtkWidget *check_header;
  GtkWidget *combo_encoding, *radio_crlf, *radio_cr, *radio_lf;
} SaveDialog;


/* ja:ラジオボタンが変更された */
static void
command_save_toggled_radio_db (GtkWidget  *widget,
                               SaveDialog *savedlg)
{
  gboolean active;

  active = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget));
  gtk_widget_set_sensitive (savedlg->frame_path, !active);
  gtk_widget_set_sensitive (savedlg->frame_sep, !active
                            && !gtk_toggle_button_get_active
                                (GTK_TOGGLE_BUTTON (savedlg->radio_basename)));
  gtk_widget_set_sensitive (savedlg->radio_crlf, !active);
  gtk_widget_set_sensitive (savedlg->radio_cr, !active);
  gtk_widget_set_sensitive (savedlg->radio_lf, !active);
  gtk_widget_set_sensitive (savedlg->check_header, !active);
}


/* ja:ラジオボタンが変更された */
static void
command_save_toggled_radio_basename (GtkWidget  *widget,
                                     SaveDialog *savedlg)
{
  gtk_widget_set_sensitive (savedlg->frame_sep,
                !gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)));
}


/* ja:コンボボックスが変更された */
static void
command_save_edited (GtkWidget *widget,
                     GtkWidget *dialog)
{
#if GTK_CHECK_VERSION(2,4,0)
  gtk_dialog_set_response_sensitive (GTK_DIALOG (dialog), GTK_RESPONSE_ACCEPT,
#else /* not GTK_CHECK_VERSION(2,4,0) */
  gtk_dialog_set_response_sensitive (GTK_DIALOG (dialog), GTK_RESPONSE_OK,
#endif /* not GTK_CHECK_VERSION(2,4,0) */
                                orz_charset_is_valid (ORZ_CHARSET (widget)));
}


void
command_saveas (GtkWidget *widget,
                gpointer   user_data)
{
  gboolean header;
  gchar *file, *charset, *title;
  gint playlist, separator, crlf, page_num;
  guint filetype;
  GtkStockItem stock_item;
  GtkWidget *dialog, *frame, *hbox, *vbox;
  GtkWidget *label_path, *label_sep, *label_encoding;
  MmaidWindow *mmaid;
  SaveDialog savedlg;

  page_num = gtk_notebook_get_current_page (GTK_NOTEBOOK (mdi));
  mmaid = orz_mdi_get_data (ORZ_MDI (mdi), page_num);
  /* ja:ダイアログ */
  gtk_stock_lookup (GTK_STOCK_SAVE_AS, &stock_item);
  title = misc_mnemonic_to_text (stock_item.label);
#if GTK_CHECK_VERSION(2,4,0)
  dialog = gtk_file_chooser_dialog_new (title,
                        GTK_WINDOW (window), GTK_FILE_CHOOSER_ACTION_SAVE,
# ifdef G_OS_WIN32
                        GTK_STOCK_SAVE_AS, GTK_RESPONSE_ACCEPT,
                        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
# else /* not G_OS_WIN32 */
                        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                        GTK_STOCK_SAVE_AS, GTK_RESPONSE_ACCEPT,
# endif /* not G_OS_WIN32 */
                        NULL);
#else /* not GTK_CHECK_VERSION(2,4,0) */
  dialog = gtk_file_selection_new (title);
#endif /* not GTK_CHECK_VERSION(2,4,0) */
  g_free (title);
  /* ja:チェックボタン */
  savedlg.check_header = gtk_check_button_new_with_mnemonic (_("_Header"));
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (savedlg.check_header),
                                                                mmaid->header);
  /* ja:ラジオボタン */
  savedlg.radio_db = gtk_radio_button_new_with_mnemonic (NULL, _("_iDB File"));
  savedlg.radio_plist = gtk_radio_button_new_with_mnemonic_from_widget
                    (GTK_RADIO_BUTTON (savedlg.radio_db), _("_Play List"));
  savedlg.radio_relative = gtk_radio_button_new_with_label
                    (NULL, _("Relative"));
  savedlg.radio_absolute = gtk_radio_button_new_with_label_from_widget
                    (GTK_RADIO_BUTTON (savedlg.radio_relative), _("Absolute"));
  savedlg.radio_basename = gtk_radio_button_new_with_label_from_widget
                    (GTK_RADIO_BUTTON (savedlg.radio_absolute), _("File Name"));

  savedlg.radio_slash = gtk_radio_button_new_with_label (NULL, "/");
  savedlg.radio_yen = gtk_radio_button_new_with_label_from_widget
                    (GTK_RADIO_BUTTON (savedlg.radio_slash), "\\");
  savedlg.radio_crlf = gtk_radio_button_new_with_label (NULL, "CR+LF");
  savedlg.radio_cr = gtk_radio_button_new_with_label_from_widget
                    (GTK_RADIO_BUTTON (savedlg.radio_crlf), "CR");
  savedlg.radio_lf = gtk_radio_button_new_with_label_from_widget
                    (GTK_RADIO_BUTTON (savedlg.radio_cr), "LF");
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON
                (mmaid->filetype == FILE_TYPE_DB ? savedlg.radio_db
                                                 : savedlg.radio_plist), TRUE);
  switch (mmaid->playlist)
    {
      case PLAYLIST_RELATIVE:
        gtk_toggle_button_set_active
                            (GTK_TOGGLE_BUTTON (savedlg.radio_relative), TRUE);
        break;
      case PLAYLIST_ABSOLUTE:
        gtk_toggle_button_set_active
                            (GTK_TOGGLE_BUTTON (savedlg.radio_absolute), TRUE);
        break;
      default:
        gtk_toggle_button_set_active
                            (GTK_TOGGLE_BUTTON (savedlg.radio_basename), TRUE);
    }
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON
                        (mmaid->separator == '/' ? savedlg.radio_slash
                                                 : savedlg.radio_yen), TRUE);
  switch (mmaid->crlf)
    {
      case CODE_CRLF:
        gtk_toggle_button_set_active
                                (GTK_TOGGLE_BUTTON (savedlg.radio_crlf), TRUE);
        break;
      case CODE_CR:
        gtk_toggle_button_set_active
                                (GTK_TOGGLE_BUTTON (savedlg.radio_cr), TRUE);
        break;
      default:
        gtk_toggle_button_set_active
                                (GTK_TOGGLE_BUTTON (savedlg.radio_lf), TRUE);
    }
  g_signal_connect (G_OBJECT (savedlg.radio_db), "toggled",
                G_CALLBACK (command_save_toggled_radio_db), &savedlg);
  g_signal_connect (G_OBJECT (savedlg.radio_basename), "toggled",
                G_CALLBACK (command_save_toggled_radio_basename), &savedlg);
  /* ja:文字符号化方式 */
  savedlg.combo_encoding = orz_charset_new ();
  orz_charset_add_candidate (ORZ_CHARSET (savedlg.combo_encoding),
                                                                charset_list);
  orz_charset_set_charset (ORZ_CHARSET (savedlg.combo_encoding),
                                                            mmaid->charset);
  g_signal_connect (G_OBJECT (savedlg.combo_encoding), "edited",
                                    G_CALLBACK (command_save_edited), dialog);
  /* ja:ラベル */
  label_path = gtk_label_new_with_mnemonic (_("_Path"));
  gtk_label_set_mnemonic_widget (GTK_LABEL (label_path),
                                                    savedlg.radio_relative);
  label_sep = gtk_label_new_with_mnemonic (_("_Separator"));
  gtk_label_set_mnemonic_widget (GTK_LABEL (label_sep), savedlg.radio_slash);
  label_encoding = gtk_label_new_with_mnemonic (_("Character _Encoding"));
  gtk_label_set_mnemonic_widget (GTK_LABEL (label_encoding),
                                                    savedlg.combo_encoding);
  /* ja:フレームとボックス */
  vbox = gtk_vbox_new (FALSE, SPACING);
  gtk_container_set_border_width (GTK_CONTAINER (vbox), SPACING);
  savedlg.frame_path = gtk_frame_new (NULL);
  gtk_frame_set_label_widget (GTK_FRAME (savedlg.frame_path), label_path);
  hbox = gtk_hbox_new (FALSE, SPACING);
  gtk_container_set_border_width (GTK_CONTAINER (hbox), SPACING);
  gtk_box_pack_start (GTK_BOX (hbox), savedlg.radio_relative, FALSE, FALSE, 0);
  gtk_box_pack_start (GTK_BOX (hbox), savedlg.radio_absolute, FALSE, FALSE, 0);
  gtk_box_pack_start (GTK_BOX (hbox), savedlg.radio_basename, FALSE, FALSE, 0);
  gtk_container_add (GTK_CONTAINER (savedlg.frame_path), hbox);
  savedlg.frame_sep = gtk_frame_new (NULL);
  gtk_frame_set_label_widget (GTK_FRAME (savedlg.frame_sep), label_sep);
  hbox = gtk_hbox_new (FALSE, SPACING);
  gtk_container_set_border_width (GTK_CONTAINER (hbox), SPACING);
  gtk_box_pack_start (GTK_BOX (hbox), savedlg.radio_slash, FALSE, FALSE, 0);
  gtk_box_pack_start (GTK_BOX (hbox), savedlg.radio_yen, FALSE, FALSE, 0);
  gtk_container_add (GTK_CONTAINER (savedlg.frame_sep), hbox);
  hbox = gtk_hbox_new (FALSE, SPACING);
  gtk_box_pack_start (GTK_BOX (hbox), savedlg.radio_db, FALSE, FALSE, 0);
  gtk_box_pack_start (GTK_BOX (hbox), savedlg.radio_plist, FALSE, FALSE, 0);
  gtk_box_pack_start (GTK_BOX (hbox), savedlg.frame_path, FALSE, FALSE, 0);
  gtk_box_pack_start (GTK_BOX (hbox), savedlg.frame_sep, FALSE, FALSE, 0);
  gtk_box_pack_start (GTK_BOX (hbox), savedlg.check_header, FALSE, FALSE, 0);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
  frame = gtk_frame_new (NULL);
  gtk_frame_set_label_widget (GTK_FRAME (frame), label_encoding);
  hbox = gtk_hbox_new (FALSE, SPACING);
  gtk_container_set_border_width (GTK_CONTAINER (hbox), SPACING);
  gtk_box_pack_start (GTK_BOX (hbox), savedlg.combo_encoding, TRUE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (hbox), savedlg.radio_crlf, FALSE, FALSE, 0);
  gtk_box_pack_start (GTK_BOX (hbox), savedlg.radio_cr, FALSE, FALSE, 0);
  gtk_box_pack_start (GTK_BOX (hbox), savedlg.radio_lf, FALSE, FALSE, 0);
  gtk_container_add (GTK_CONTAINER (frame), hbox);
  gtk_box_pack_start (GTK_BOX (vbox), frame, TRUE, TRUE, 0);
  /* ja:表示 */
  command_save_toggled_radio_db (savedlg.radio_db, &savedlg);
  gtk_widget_show_all (vbox);
  /* ja:ダイアログ */
#if GTK_CHECK_VERSION(2,4,0)
  gtk_file_chooser_set_extra_widget (GTK_FILE_CHOOSER (dialog), vbox);
  if (save_path)
    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (dialog), save_path);
  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    {
# if ! GTK_CHECK_VERSION(2,6,0)
      GtkTreeIter iter;
# endif /* not GTK_CHECK_VERSION(2,6,0) */

      file = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
#else /* not GTK_CHECK_VERSION(2,4,0) */
  gtk_box_pack_end (GTK_BOX (GTK_FILE_SELECTION (dialog)->main_vbox),
                                                        vbox, FALSE, FALSE, 0);
  if (save_path)
    {
      gchar *path;

      path = g_strconcat (save_path, G_SEARCHPATH_SEPARATOR_S, NULL);
      gtk_file_selection_set_filename (GTK_FILE_SELECTION (dialog), path);
      g_free (path);
    }
  gtk_file_selection_hide_fileop_buttons (GTK_FILE_SELECTION (dialog));
  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_OK)
    {
      file = g_strdup (gtk_file_selection_get_filename
                                                (GTK_FILE_SELECTION (dialog)));
#endif /* not GTK_CHECK_VERSION(2,4,0) */
      charset = orz_charset_get_charset (ORZ_CHARSET (savedlg.combo_encoding));
      filetype = gtk_toggle_button_get_active
        (GTK_TOGGLE_BUTTON (savedlg.radio_db)) ? FILE_TYPE_DB : FILE_TYPE_LIST;
      if (gtk_toggle_button_get_active
                                (GTK_TOGGLE_BUTTON (savedlg.radio_relative)))
        playlist = PLAYLIST_RELATIVE;
      else if (gtk_toggle_button_get_active
                                (GTK_TOGGLE_BUTTON (savedlg.radio_absolute)))
        playlist = PLAYLIST_ABSOLUTE;
      else
        playlist = PLAYLIST_BASENAME;
      separator = gtk_toggle_button_get_active
                        (GTK_TOGGLE_BUTTON (savedlg.radio_slash)) ? '/' : '\\';
      if (gtk_toggle_button_get_active
                                    (GTK_TOGGLE_BUTTON (savedlg.radio_crlf)))
        crlf = CODE_CRLF;
      else if (gtk_toggle_button_get_active
                                    (GTK_TOGGLE_BUTTON (savedlg.radio_cr)))
        crlf = CODE_CR;
      else
        crlf = CODE_LF;
      header = gtk_toggle_button_get_active
                                    (GTK_TOGGLE_BUTTON (savedlg.check_header));
    }
  else
    {
      file = charset = NULL;
      filetype = FILE_TYPE_DB;
      playlist = PLAYLIST_RELATIVE;
      separator = G_DIR_SEPARATOR;
#ifdef G_OS_WIN32
      crlf = CODE_CRLF;
#else /* not G_OS_WIN32 */
      crlf = CODE_LF;
#endif /* not G_OS_WIN32 */
      header = TRUE;
    }
  gtk_widget_destroy (dialog);
  if (file && charset)
    {
      gboolean result = TRUE;

      if (g_file_test (file, G_FILE_TEST_EXISTS))
        {
          gchar *utf8str;

          utf8str = g_filename_display_name (file);
          dialog = gtk_message_dialog_new (GTK_WINDOW (window),
                    GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
                    GTK_MESSAGE_INFO, GTK_BUTTONS_YES_NO,
                    _("%s\nThis file is existed.\n\nOver write?"), utf8str);
          g_free (utf8str);
          g_signal_connect (G_OBJECT (dialog), "key-press-event",
                                    G_CALLBACK (misc_dialog_key_press), NULL);
          result = gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_YES;
          gtk_widget_destroy (dialog);
        }
      if (result && (filetype == FILE_TYPE_DB
                                    ? file_save_db (file, mmaid, charset)
                                    : file_save_list (file, mmaid, charset,
                                        playlist, separator, crlf, header)))
        {
          gchar *path;

          g_free (save_path);
          path = g_path_get_absolute (file, NULL);
          save_path = g_path_get_dirname (path);
          g_free (path);
          /* ja:キャラクターセット */
          g_free (mmaid->charset);
          mmaid->charset = g_strdup (charset);
          if (filetype == FILE_TYPE_DB || playlist != PLAYLIST_BASENAME)
            {
              /* ja:諸設定 */
              orz_mdi_set_created (ORZ_MDI (mdi), page_num, FALSE);
              orz_mdi_set_edited (ORZ_MDI (mdi), page_num, FALSE);
              orz_mdi_set_file (ORZ_MDI (mdi), page_num, file);
            }
          mmaid->filetype = filetype;
          if (filetype == FILE_TYPE_LIST)
            {
              def_plist = mmaid->playlist = playlist;
              def_sep = mmaid->separator = separator;
              def_crlf = mmaid->crlf = crlf;
              def_head = mmaid->header = header;
            }
          orz_history_add_file (ORZ_HISTORY (history), file);
        }
    }
  g_free (file);
  g_free (charset);
}


void
command_reload (GtkWidget *widget,
                gpointer   user_data)
{
  gchar *text;
  const gchar *file;
  gint page_num;
  MmaidWindow *mmaid;

  page_num = gtk_notebook_get_current_page (GTK_NOTEBOOK (mdi));
  mmaid = orz_mdi_get_data (ORZ_MDI (mdi), page_num);
  file = orz_mdi_get_file (ORZ_MDI (mdi), page_num);
  text = orz_mdi_get_edited (ORZ_MDI (mdi), page_num) ? g_strdup_printf
            (_("File %s was edited.\nRestore to existed file?"), file) : NULL;
  if (reload_dialog (mmaid, text))
    {
      /* ja:削除 */
      mm_list_delete_all (MM_LIST (mmaid->list));
      /* ja:再読み込み */
      file_open_db (file, mmaid);
      orz_mdi_set_edited (ORZ_MDI (mdi), page_num, FALSE);
      charlist_renewal_all (mmaid->charset);
      set_menu_bar (mmaid);
    }
}


static void
command_dialog_abort (GtkWidget *widget,
                      GtkWidget *dialog)
{
  userbreak = FALSE;
}


void
command_export (GtkWidget *widget,
                gpointer   user_data)
{
  gint i, count;
  GtkWidget *dialog;
  ExportInfo *expinfo;
  MmaidWindow *mmaid;

  mmaid = orz_mdi_get_data (ORZ_MDI (mdi), -1);
  expinfo = export_dialog (mmaid->charset);
  if (!expinfo)
    return;
  def_hold = expinfo->hold;
  def_gnum = expinfo->gnum;
  userbreak = TRUE;
  dialog = mm_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (dialog), _("Export"));
  mm_list_set_editable (MM_LIST (MM_DIALOG (dialog)->list), TRUE);
  g_signal_connect (G_OBJECT (dialog), "destroy",
                                            G_CALLBACK (gtk_main_quit), NULL);
  g_signal_connect (G_OBJECT (MM_DIALOG (dialog)->abort_button),
                        "clicked", G_CALLBACK (command_dialog_abort), dialog);
  g_signal_connect_swapped (G_OBJECT (MM_DIALOG (dialog)->ok_button),
                "clicked", G_CALLBACK (gtk_widget_destroy), G_OBJECT (dialog));
  g_object_set_data (G_OBJECT (dialog), "user_data", mmaid);
  gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (window));
  gtk_widget_show (MM_DIALOG (dialog)->abort_button);
  gtk_widget_show (dialog);
  gtk_grab_add (dialog);
  count = mm_list_length (MM_LIST (mmaid->list));
  for (i = 0; i < count && userbreak; i++)
    {
      while (gtk_events_pending ())
        gtk_main_iteration ();
      if (mm_list_is_selected (MM_LIST (mmaid->list), i))
        {
          gchar *file, *str;
          gint j, k, n, gnum;
          ID3Tag *id3tag;

          id3tag = copy_id3tag (mm_list_get_nth (MM_LIST (mmaid->list), i));
          file = g_filename_from_utf8 (id3tag->file, -1, NULL, NULL, NULL);
          switch (file_is_type (file))
            {
              case FILE_TYPE_ERROR:
                /* ja:ファイルが存在しないとき */
                str = _("File Open Error");
                break;
              case FILE_TYPE_TAG:
                if (expinfo->hold)
                  {
                    /* ja:既存のタグを保持し、かつ既にタグがあるとき */
                    str = _("Have Tag");
                    break;
                  }
              default:
                gnum = -1;
                if (id3tag->genre)
                  for (j = 0; j < genres; j++)
                    for (k = 0; k < 256; k++)
                      if (genre[j][k]
                                    && !g_strcmp (genre[j][k], id3tag->genre))
                        {
                          gnum = k;
                          goto loop;
                        }
                loop:
                if (gnum == -1)
                  {
                    gnum = expinfo->gnum;
                    g_free (id3tag->genre);
                    id3tag->genre = NULL;
                    for (j = 0; j < genres; j++)
                      if (genre[j][gnum])
                        {
                          id3tag->genre = g_strdup (genre[j][gnum]);
                          break;
                        }
                  }
                if (!id3tag->charset)
                  id3tag->charset = g_strdup (expinfo->charset);
                switch (file_save_id3tag (file, id3tag, gnum))
                  {
                    case FILE_SAVE_SUCCESS:
                      str = _("Success");
                      break;
                    case FILE_SAVE_CHARSET_TITLE:
                      str = _("Invalid Character Encoding of Title");
                      break;
                    case FILE_SAVE_CHARSET_ARTIST:
                      str = _("Invalid Character Encoding of Artist");
                      break;
                    case FILE_SAVE_CHARSET_ALBUM:
                      str = _("Invalid Character Encoding of Album");
                      break;
                    case FILE_SAVE_CHARSET_YEAR:
                      str = _("Invalid Character Encoding of Year");
                      break;
                    case FILE_SAVE_CHARSET_COMMENT:
                      str = _("Invalid Character Encoding of Comment");
                      break;
                    case FILE_SAVE_LENGTH_TITLE:
                      str = _("Invalid Length of Title");
                      break;
                    case FILE_SAVE_LENGTH_ARTIST:
                      str = _("Invalid Length of Artist");
                      break;
                    case FILE_SAVE_LENGTH_ALBUM:
                      str = _("Invalid Length of Album");
                      break;
                    case FILE_SAVE_LENGTH_YEAR:
                      str = _("Invalid Length of Year");
                      break;
                    case FILE_SAVE_LENGTH_COMMENT:
                      str = _("Invalid Length of Comment");
                      break;
                    default:
                      str = _("Error");
                  }
            }
          g_free (file);
          n = mm_list_append (MM_LIST (MM_DIALOG (dialog)->list), id3tag);
          mm_list_set_status (MM_LIST (MM_DIALOG (dialog)->list), n, str);
        }
    }
  g_free (expinfo->charset);
  g_free (expinfo);
  gtk_widget_hide (MM_DIALOG (dialog)->abort_button);
  gtk_widget_show (MM_DIALOG (dialog)->ok_button);
  gtk_main ();
}


void
command_exit (GtkWidget *widget,
              gpointer   user_data)
{
  GdkEvent event;

  event.any.type = GDK_DELETE;
  event.any.window = gtk_widget_get_window (window);
  event.any.send_event = FALSE;
  gdk_event_put (&event);
}


void
command_up (GtkWidget *widget,
            gpointer   user_data)
{
  gint i, count, page_num;
  MmaidWindow *mmaid;

  page_num = gtk_notebook_get_current_page (GTK_NOTEBOOK (mdi));
  mmaid = orz_mdi_get_data (ORZ_MDI (mdi), page_num);
  count = mm_list_length (MM_LIST (mmaid->list));
  for (i = 1; i < count; i++)
    if (!mm_list_is_selected (MM_LIST (mmaid->list), i - 1)
                            && mm_list_is_selected (MM_LIST (mmaid->list), i))
      mm_list_swap (MM_LIST (mmaid->list), i - 1, i);
  orz_mdi_set_edited (ORZ_MDI (mdi), page_num, TRUE);
}


void
command_down (GtkWidget *widget,
              gpointer   user_data)
{
  gint i, page_num;
  MmaidWindow *mmaid;

  page_num = gtk_notebook_get_current_page (GTK_NOTEBOOK (mdi));
  mmaid = orz_mdi_get_data (ORZ_MDI (mdi), page_num);
  for (i = mm_list_length (MM_LIST (mmaid->list)) - 2; i >= 0 ; i--)
    if (mm_list_is_selected (MM_LIST (mmaid->list), i)
                        && !mm_list_is_selected (MM_LIST (mmaid->list), i + 1))
      mm_list_swap (MM_LIST (mmaid->list), i, i + 1);
  orz_mdi_set_edited (ORZ_MDI (mdi), page_num, TRUE);
}


void
command_root (GtkWidget *widget,
              gpointer   user_data)
{
  root_dialog ((MmaidWindow *)orz_mdi_get_data (ORZ_MDI (mdi), -1));
}


static void
command_dialog_ok (GtkWidget *widget,
                   GtkWidget *dialog)
{
  gboolean edited = FALSE;
  gint i, count;
  MmaidWindow *mmaid;

  mmaid = g_object_get_data (G_OBJECT (dialog), "user_data");
  count = mm_list_length (MM_LIST (MM_DIALOG (dialog)->list));
  for (i = 0; i < count; i++)
    {
      gint n;
      const ID3Tag *id3tag;

      id3tag = mm_list_get_nth (MM_LIST (MM_DIALOG (dialog)->list), i);
      n = mm_list_search (MM_LIST (mmaid->list),
                                            MM_LIST_FILE, id3tag->file, -1);
      if (n >= 0)
        {
          if (mm_list_get_active_refresh
                                    (MM_LIST (MM_DIALOG (dialog)->list), i))
            {
              mm_list_change (MM_LIST (mmaid->list), n,
                                    MM_LIST_ID3TAG, copy_id3tag (id3tag), -1);
              edited = TRUE;
            }
          else if (mm_list_get_active_delete
                                    (MM_LIST (MM_DIALOG (dialog)->list), i))
            {
              mm_list_delete (MM_LIST (mmaid->list), n);
              edited = TRUE;
            }
        }
      else
        {
          if (mm_list_get_active_add (MM_LIST (MM_DIALOG (dialog)->list), i))
            {
              mm_list_append (MM_LIST (mmaid->list), copy_id3tag (id3tag));
              edited = TRUE;
            }
        }
    }
  if (edited)
    orz_mdi_set_edited (ORZ_MDI (mdi),
                    orz_mdi_get_page_from_data (ORZ_MDI (mdi), mmaid), TRUE);
  gtk_widget_destroy (dialog);
}


void
command_refresh (GtkWidget *widget,
                 gpointer   user_data)
{
  gboolean selected = FALSE;
  gchar *title;
  gint i, count;
  GtkStockItem stock_item;
  GtkWidget *dialog;
  MmaidWindow *mmaid;

  mmaid = orz_mdi_get_data (ORZ_MDI (mdi), -1);
  /* ja:選択の数を数える */
  count = mm_list_length (MM_LIST (mmaid->list));
  for (i = 0; i < count; i++)
    if (mm_list_is_selected (MM_LIST (mmaid->list), i))
      {
        selected = TRUE;
        break;
      }
  userbreak = TRUE;
  dialog = mm_dialog_new ();
  gtk_stock_lookup (GTK_STOCK_REFRESH, &stock_item);
  title = misc_mnemonic_to_text (stock_item.label);
  gtk_window_set_title (GTK_WINDOW (dialog), title);
  g_free (title);
  mm_list_set_editable (MM_LIST (MM_DIALOG (dialog)->list), TRUE);
  g_signal_connect (G_OBJECT (MM_DIALOG (dialog)->abort_button),
                        "clicked", G_CALLBACK (command_dialog_abort), dialog);
  g_signal_connect (G_OBJECT (MM_DIALOG (dialog)->ok_button),
                        "clicked", G_CALLBACK (command_dialog_ok), dialog);
  g_signal_connect_swapped (G_OBJECT (MM_DIALOG (dialog)->cancel_button),
                "clicked", G_CALLBACK (gtk_widget_destroy), G_OBJECT (dialog));
  g_object_set_data (G_OBJECT (dialog), "user_data", mmaid);
  gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (window));
  gtk_widget_show (MM_DIALOG (dialog)->abort_button);
  gtk_widget_show (dialog);
  gtk_grab_add (dialog);
  for (i = 0; i < count && userbreak; i++)
    {
      while (gtk_events_pending ())
        gtk_main_iteration ();
      if (!selected || mm_list_is_selected (MM_LIST (mmaid->list), i))
        {
          gchar *file;
          gint n;
          const ID3Tag *id3tag0;
          ID3Tag *id3tag1;

          id3tag0 = mm_list_get_nth (MM_LIST (mmaid->list), i);
          file = g_filename_from_utf8 (id3tag0->file, -1, NULL, NULL, NULL);
          switch (file_is_type (file))
            {
              case FILE_TYPE_MUSIC:
                id3tag1 = file_open_id3tag (file, id3tag0->charset
                                            ? id3tag0->charset : charset_list);
                n = mm_list_append (MM_LIST (MM_DIALOG (dialog)->list),
                                                                    id3tag1);
                if (!cmp_id3tag (id3tag0, id3tag1))
                  mm_list_show_refresh (MM_LIST (MM_DIALOG (dialog)->list), n);
                mm_list_show_delete (MM_LIST (MM_DIALOG (dialog)->list), n);
                break;
              case FILE_TYPE_TAG:
                id3tag1 = file_open_id3tag (file, id3tag0->charset
                                            ? id3tag0->charset : charset_list);
                n = mm_list_append (MM_LIST (MM_DIALOG (dialog)->list),
                                                                    id3tag1);
                mm_list_show_delete (MM_LIST (MM_DIALOG (dialog)->list), n);
                if (!cmp_id3tag (id3tag0, id3tag1))
                  {
                    mm_list_set_sensitive_delete
                                (MM_LIST (MM_DIALOG (dialog)->list), n, FALSE);
                    mm_list_show_refresh (MM_LIST (MM_DIALOG (dialog)->list),
                                                                            n);
                    mm_list_set_active_refresh
                                (MM_LIST (MM_DIALOG (dialog)->list), n, TRUE);
                  }
                break;
              default:
                n = mm_list_append (MM_LIST (MM_DIALOG (dialog)->list),
                                                        copy_id3tag (id3tag0));
                mm_list_show_delete (MM_LIST (MM_DIALOG (dialog)->list), n);
                mm_list_set_active_delete (MM_LIST (MM_DIALOG (dialog)->list),
                                                                    n, TRUE);
                break;
            }
          g_free (file);
        }
    }
  if (userbreak)
    {
      gtk_widget_hide (MM_DIALOG (dialog)->abort_button);
      gtk_widget_show (MM_DIALOG (dialog)->ok_button);
      gtk_widget_show (MM_DIALOG (dialog)->cancel_button);
      g_signal_connect (G_OBJECT (dialog), "destroy",
                                            G_CALLBACK (gtk_main_quit), NULL);
      gtk_main ();
    }
  else
    {
      gtk_widget_destroy (dialog);
    }
}


void
command_add (GtkWidget *widget,
             gpointer   user_data)
{
  add_dialog ((MmaidWindow *)orz_mdi_get_data (ORZ_MDI (mdi), -1));
}


void
command_edit (GtkWidget *widget,
              gpointer   user_data)
{
  edit_dialog ((MmaidWindow *)orz_mdi_get_data (ORZ_MDI (mdi), -1));
}


static void
command_search_recursive (const gchar *path,
                          GtkWidget   *dialog)
{
  GDir *gdir;

  gdir = g_dir_open (path, 0, NULL);
  if (gdir)
    {
      const gchar *name;

      while (userbreak && (name = g_dir_read_name (gdir)))
        {
          gchar *file;

          while (gtk_events_pending ())
            gtk_main_iteration ();
#ifndef G_OS_WIN32
          if (name[0] == '.')
            continue;
#endif /* not G_OS_WIN32 */
          file = g_build_filename (path, name, NULL);
          if (g_file_test (file, G_FILE_TEST_IS_DIR))
            {
              command_search_recursive (file, dialog);
            }
          else
            {
              gchar *utf8file;
              gint n0, n1;
              const ID3Tag *id3tag0;
              ID3Tag *id3tag1;
              MmaidWindow *mmaid;

              mmaid = g_object_get_data (G_OBJECT (dialog), "user_data");
              utf8file = g_filename_to_utf8 (file, -1, NULL, NULL, NULL);
              n0 = mm_list_search (MM_LIST (mmaid->list),
                                                MM_LIST_FILE, utf8file, -1);
              g_free (utf8file);
              id3tag0 = n0 >= 0 ? mm_list_get_nth (MM_LIST (mmaid->list), n0)
                                : NULL;
              switch (file_is_type (file))
                {
                  case FILE_TYPE_MUSIC:
                    id3tag1 = file_open_id3tag (file, charset_list);
                    n1 = mm_list_append (MM_LIST (MM_DIALOG (dialog)->list),
                                                                    id3tag1);
                    if (n0 >= 0)
                      {
                        if (!cmp_id3tag (id3tag0, id3tag1))
                          mm_list_show_refresh
                                (MM_LIST (MM_DIALOG (dialog)->list), n1);
                        mm_list_show_delete
                                (MM_LIST (MM_DIALOG (dialog)->list), n1);
                      }
                    else
                      {
                        mm_list_show_add
                                (MM_LIST (MM_DIALOG (dialog)->list), n1);
                        if (g_strlen (file) > 4 && g_ascii_strcasecmp (".mp3",
                                            file + g_strlen (file) - 4) == 0)
                          mm_list_set_active_add
                                (MM_LIST (MM_DIALOG (dialog)->list), n1, TRUE);
                      }
                    break;
                  case FILE_TYPE_TAG:
                    id3tag1 = file_open_id3tag (file, charset_list);
                    n1 = mm_list_append (MM_LIST (MM_DIALOG (dialog)->list),
                                                                    id3tag1);
                    if (n0 >= 0)
                      {
                        mm_list_show_delete
                                (MM_LIST (MM_DIALOG (dialog)->list), n1);
                        if (!cmp_id3tag (id3tag0, id3tag1))
                          {
                            mm_list_set_sensitive_delete
                                (MM_LIST (MM_DIALOG (dialog)->list),
                                                                    n1, FALSE);
                            mm_list_show_refresh
                                (MM_LIST (MM_DIALOG (dialog)->list), n1);
                            mm_list_set_active_refresh
                                (MM_LIST (MM_DIALOG (dialog)->list), n1, TRUE);
                          }
                      }
                    else
                      {
                        mm_list_show_add
                                (MM_LIST (MM_DIALOG (dialog)->list), n1);
                        mm_list_set_active_add
                                (MM_LIST (MM_DIALOG (dialog)->list), n1, TRUE);
                      }
                    break;
                  default:
                    if (n0 >= 0)
                      {
                        n1 = mm_list_append
                                (MM_LIST (MM_DIALOG (dialog)->list),
                                                        copy_id3tag (id3tag0));
                        mm_list_show_delete
                                (MM_LIST (MM_DIALOG (dialog)->list), n1);
                        mm_list_set_active_delete
                                (MM_LIST (MM_DIALOG (dialog)->list), n1, TRUE);
                      }
                }
            }
          g_free (file);
        }
      g_dir_close (gdir);
    }
}


void
command_search (GtkWidget *widget,
                gpointer   user_data)
{
  gchar *path, *tmp;
  const gchar *file;
  gint page_num;
  GtkWidget *dialog;
  MmaidWindow *mmaid;

  page_num = gtk_notebook_get_current_page (GTK_NOTEBOOK (mdi));
  mmaid = orz_mdi_get_data (ORZ_MDI (mdi), page_num);
  file = orz_mdi_get_file (ORZ_MDI (mdi), page_num);
  path = g_path_is_absolute (file) ? g_path_get_dirname (file)
       : open_path ? g_strdup (open_path)
       : save_path ? g_strdup (save_path)
       : NULL;
#if GTK_CHECK_VERSION(2,4,0)
  dialog = gtk_file_chooser_dialog_new (_("Browse"),
                        GTK_WINDOW (window),
                        GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER,
# ifdef G_OS_WIN32
                        GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
                        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
# else /* not G_OS_WIN32 */
                        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                        GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
# endif /* not G_OS_WIN32 */
                        NULL);
  if (path)
    {
      gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (dialog), path);
      g_free (path);
      path = NULL;
    }
  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    path = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
#else /* not GTK_CHECK_VERSION(2,4,0) */
  dialog = gtk_file_selection_new (_("Browse"));
  if (path)
    {
      if (path[0] != '\0')
        {
          gchar *tmp;

          if (!G_IS_DIR_SEPARATOR (path[g_strlen (path) - 1]))
            {
              tmp = g_strconcat (path, G_DIR_SEPARATOR_S, NULL);
              g_free (path);
              path = tmp;
            }
          gtk_file_selection_set_filename (GTK_FILE_SELECTION (dialog), path);
        }
      g_free (path);
      path = NULL;
    }
  gtk_file_selection_hide_fileop_buttons (GTK_FILE_SELECTION (dialog));
  gtk_widget_hide
            (gtk_widget_get_parent (GTK_FILE_SELECTION (dialog)->file_list));
  gtk_widget_hide (GTK_FILE_SELECTION (dialog)->selection_entry);
  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_OK)
    path = g_strdup (gtk_file_selection_get_filename
                                                (GTK_FILE_SELECTION (dialog)));
#endif /* not GTK_CHECK_VERSION(2,4,0) */
  gtk_widget_destroy (dialog);

  if (!path)
    return;
  tmp = g_filename_to_utf8 (path, -1, NULL, NULL, NULL);
  g_free (path);
  if (!tmp || tmp[0] =='\0')
    {
      g_free (tmp);
      return;
    }
  if (G_IS_DIR_SEPARATOR (tmp[g_strlen (tmp) - 1]))
    tmp[g_strlen (tmp) - 1] = '\0';
  path = g_filename_from_utf8 (tmp, -1, NULL, NULL, NULL);
  g_free (tmp);
  if (!path)
    return;
  userbreak = TRUE;
  dialog = mm_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (dialog), _("Search"));
  mm_list_set_editable (MM_LIST (MM_DIALOG (dialog)->list), TRUE);
  g_signal_connect (G_OBJECT (dialog), "destroy",
                                            G_CALLBACK (gtk_main_quit), NULL);
  g_signal_connect (G_OBJECT (MM_DIALOG (dialog)->abort_button),
                        "clicked", G_CALLBACK (command_dialog_abort), dialog);
  g_signal_connect (G_OBJECT (MM_DIALOG (dialog)->ok_button),
                        "clicked", G_CALLBACK (command_dialog_ok), dialog);
  g_signal_connect_swapped (G_OBJECT (MM_DIALOG (dialog)->cancel_button),
                "clicked", G_CALLBACK (gtk_widget_destroy), G_OBJECT (dialog));
  g_object_set_data (G_OBJECT (dialog), "user_data", mmaid);
  gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (window));
  gtk_widget_show (MM_DIALOG (dialog)->abort_button);
  gtk_widget_show (dialog);
  gtk_grab_add (dialog);
  command_search_recursive (path, dialog);
  g_free (path);
  gtk_widget_hide (MM_DIALOG (dialog)->abort_button);
  gtk_widget_show (MM_DIALOG (dialog)->ok_button);
  gtk_widget_show (MM_DIALOG (dialog)->cancel_button);
  gtk_main ();
}


void
command_cut (GtkWidget *widget,
             gpointer   user_data)
{
  command_copy (widget, user_data);
  command_delete (widget, user_data);
}


void
command_copy (GtkWidget *widget,
              gpointer   user_data)
{
  GtkTargetEntry targets[1] = {{"Melody maid ID3 Tag", 0, TARGET_MMAID}};
  gchar *clipboard_id3tag = NULL;
  gint i, count;
  MmaidWindow *mmaid;

  mmaid = orz_mdi_get_data (ORZ_MDI (mdi), -1);
  count = mm_list_length (MM_LIST (mmaid->list));
  for (i = 0; i < count; i++)
    if (mm_list_is_selected (MM_LIST (mmaid->list), i))
      {
        gchar *str;
        const ID3Tag *id3tag;

        id3tag = mm_list_get_nth (MM_LIST (mmaid->list), i);
        str = g_strdup_printf ("%s\t%s\t%s\t%s\t%s\t%s\t%d\t%s\t%s",
                                    id3tag->file    ? id3tag->file     : "",
                                    id3tag->title   ? id3tag->title    : "",
                                    id3tag->artist  ? id3tag->artist   : "",
                                    id3tag->album   ? id3tag->album    : "",
                                    id3tag->year    ? id3tag->year     : "",
                                    id3tag->comment ? id3tag->comment  : "",
                                    id3tag->track,
                                    id3tag->genre   ? id3tag->genre    : "",
                                    id3tag->charset ? id3tag->charset  : "");
        if (clipboard_id3tag)
          {
            gchar *tmp;

            tmp = g_strconcat (clipboard_id3tag, "\n", str, NULL);
            g_free (clipboard_id3tag);
            g_free (str);
            clipboard_id3tag = tmp;
          }
        else
          {
            clipboard_id3tag = str;
          }
      }
  gtk_clipboard_set_with_data (gtk_clipboard_get (GDK_SELECTION_CLIPBOARD),
                               targets,
                               G_N_ELEMENTS (targets),
                               signal_clipboard_get,
                               signal_clipboard_clear,
                               clipboard_id3tag);
}


void
command_paste (GtkWidget *widget,
               gpointer   user_data)
{
  gtk_clipboard_request_contents (gtk_clipboard_get (GDK_SELECTION_CLIPBOARD),
                                atom_id3tag, signal_clipboard_received, NULL);
}


void
command_delete (GtkWidget *widget,
                gpointer   user_data)
{
  gint i, page_num;
  MmaidWindow *mmaid;

  page_num = gtk_notebook_get_current_page (GTK_NOTEBOOK (mdi));
  mmaid = orz_mdi_get_data (ORZ_MDI (mdi), page_num);
  for (i = mm_list_length (MM_LIST (mmaid->list)) - 1; i >= 0 ; i--)
    if (mm_list_is_selected (MM_LIST (mmaid->list), i))
      mm_list_delete (MM_LIST (mmaid->list), i);
  orz_mdi_set_edited (ORZ_MDI (mdi), page_num, TRUE);
}


void
command_all (GtkWidget *widget,
             gpointer   user_data)
{
  gint i;
  MmaidWindow *mmaid;

  mmaid = orz_mdi_get_data (ORZ_MDI (mdi), -1);
  for (i = mm_list_length (MM_LIST (mmaid->list)) - 1; i >= 0 ; i--)
    if (!mm_list_is_selected (MM_LIST (mmaid->list), i))
      break;
  if (i >= 0)
    mm_list_select_all (MM_LIST (mmaid->list));
  else
    mm_list_unselect_all (MM_LIST (mmaid->list));
}


void
command_pref (GtkWidget *widget,
              gpointer   user_data)
{
  GtkWidget *dialog;

  dialog = orz_pref_new ();
  gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (window));
  orz_pref_set_history (ORZ_PREF (dialog),
                                orz_history_get_num (ORZ_HISTORY (history)));
  orz_pref_set_newfile (ORZ_PREF (dialog), newfile);
  orz_pref_set_second (ORZ_PREF (dialog), second);
  orz_pref_set_size (ORZ_PREF (dialog), def_width, def_height);
  orz_pref_set_tab (ORZ_PREF (dialog), n_pos);
  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_OK)
    {
      orz_history_set_num (ORZ_HISTORY (history),
                                    orz_pref_get_history (ORZ_PREF (dialog)));
      newfile = orz_pref_get_newfile (ORZ_PREF (dialog));
      second = orz_pref_get_second (ORZ_PREF (dialog));
      orz_pref_get_size (ORZ_PREF (dialog), &def_width, &def_height);
      n_pos = orz_pref_get_tab (ORZ_PREF (dialog));
      /* ja:ノートの位置 */
      gtk_notebook_set_tab_pos (GTK_NOTEBOOK (mdi), n_pos);
    }
  gtk_widget_destroy (dialog);
}


void
command_genre (GtkWidget *widget,
               gpointer   user_data)
{
  genre_dialog ();
}


void
command_charset (GtkWidget *widget,
                 gpointer   user_data)
{
  GtkWidget *dialog;

  dialog = orz_charlist_new ();
  gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (window));
  orz_charlist_set_charlist (ORZ_CHARLIST (dialog), charset_list);
  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_OK)
    {
      g_free (charset_list);
      charset_list = orz_charlist_get_charlist (ORZ_CHARLIST (dialog));
      charlist_renewal_all (charset_list);
    }
  gtk_widget_destroy (dialog);
}


void
command_process (GtkWidget *widget,
                 gpointer   user_data)
{
  process_dialog ();
}


void
command_version (GtkWidget *widget,
                 gpointer   user_data)
{
  version_dialog ();
}


void
command_play (GtkWidget *widget,
              gpointer   user_data)
{
  gchar *cmd;

  cmd = g_filename_from_utf8 (cmd_play, -1, NULL, NULL, NULL);
  if (cmd)
    {
      gint n;
      MmaidWindow *mmaid;

      mmaid = orz_mdi_get_data (ORZ_MDI (mdi), -1);
      mm_list_get_cursor (MM_LIST (mmaid->list), &n, NULL);
      if (n >= 0)
        {
          gchar *file;
          const ID3Tag *id3tag;

          id3tag = mm_list_get_nth (MM_LIST (mmaid->list), n);
          file = g_filename_from_utf8 (id3tag->file, -1, NULL, NULL, NULL);
          if (file)
            {
              gchar *quoted, *str;

              quoted = g_shell_quote (file);
              str = g_strconcat (cmd, " ", quoted, NULL);
              if (quoted && str)
                g_spawn_command_line_async (str, NULL);
              g_free (str);
              g_free (quoted);
              g_free (file);
            }
        }
      g_free (cmd);
    }
}


void
command_stop (GtkWidget *widget,
              gpointer   user_data)
{
  gchar *cmd;

  cmd = g_filename_from_utf8 (cmd_stop, -1, NULL, NULL, NULL);
  if (cmd)
    {
      g_spawn_command_line_async (cmd, NULL);
      g_free (cmd);
    }
}
